define([
    'dojo/_base/lang',
    'dojo/_base/array',
    'dojo/aspect',
    'dojo/dom-construct',
    'dojo/dom-style',
    'dojo/dom-prop',
    'dojo/dom-geometry',
    'dojo/on',
    'dojo/string',
    'dojo/query',
    'dojo/date/locale',
    'dojo/request/notify',
    'dojo/query',
    'STM/common/util/ApplicationConstants'
], function(lang, array, aspect, domConstruct, domStyle, domProp, domGeom, on, string, query, locale, notify, query, AppConst) {

    // The STM date pattern used to pass dates to most of the services
    var STMDateFormat = {
        selector: 'date',
        datePattern: 'MMddyyy'
    };

    var Util = {
        // This will set the value of form widgets without firing the onChange event
        formSetNoChange: function(formWidget, valueProp, value) {
            formWidget._onChangeActive = false;
            formWidget.set(valueProp || 'value', value, true);
            formWidget._onChangeActive = true;
        },
        // Is a string effectively blank?
        isBlank : function(value) {
            return value === null || value === undefined || string.trim(value).length === 0;
        },
        // Apparently the global isNaN is flawed in javascript, and this check is superior
        isNaN: function(value) {
            return value != value;
        },
        // An equality check that counts two NaN values as equal
        equalNaN: function(first, second) {
            return first === second || (first != first && second != second);
        },
        // Most reliable way to check for an array
        isArray: function(testArray) {
            return Object.prototype.toString.call(testArray) === '[object Array]';
        },
        isNonEmptyArray: function(testArray) {
            return Util.isArray(testArray) && testArray.length > 0;
        },
        // Returns true for falsy values (except 0), empty strings, and empty arrays
        isEmpty: function(value) {
            if (!value && value !== 0) {
                return true;
            }
            if (value instanceof Array && value.length <= 0) {
                return true;
            }

            return false;
        },
        getKeyByValue: function(object, value) {
            for (var property in object) {
                if (object.hasOwnProperty(property)) {
                    if (object[property] === value) {
                        return property;
                    }
                }
            }
        },
        toSTMDateFormat: function(dateObject) {
            return locale.format(dateObject, STMDateFormat);
        },
        fromSTMDateFormat: function(dateString) {
            return locale.parse(dateString, STMDateFormat);
        },
        // Use this function to destroyRecursive() a widget if it could possibly already be destroyed
        safeDestroyRecursive: function(widget) {
            if ('destroyRecursive' in widget) {
                widget.destroyRecursive();
            }
        },
        // This will setup the busy cursor so that it will show whenever a request to the server is being made
        // while any asynchronous request is in progress. "node" should indicate the top level node of the app
        setupBusyCursor: function(appNode) {
            var showHide = function(appNode, show) {
                query('*', appNode).forEach(function(node) {
                    domStyle.set(node, 'cursor', show ? 'wait' : '');
                });
            };

            notify('send', function(response, cancel) {
                showHide(appNode, true);
            });

            notify('load', function(response) {
                showHide(appNode, false);
            });

            notify('error', function(response) {
                showHide(appNode, false);
            });

            // Ensure this function only runs once
            this.setupBusyCursor = function() {};
        },
        setupTimeout: function() {
            notify('done', function(responseOrError) {
                Util.extendPortalSession();
            });
        },
        // Setup any "shims" aka patch in features missing in older browsers
        setupShims: function() {

            // getComputedStyle
            if (!window.getComputedStyle) {
                window.getComputedStyle = function(el, pseudo) {
                    this.el = el;
                    this.getPropertyValue = function(prop) {
                        var re = /(\-([a-z]){1})/g;
                        if (prop == 'float') prop = 'styleFloat';
                        if (re.test(prop)) {
                            prop = prop.replace(re, function () {
                                return arguments[2].toUpperCase();
                            });
                        }
                        return el.currentStyle[prop] ? el.currentStyle[prop] : null;
                    }
                    return this;
                }
            }

        },
        extendPortalSession: function() {
            if (window.top.sas_framework_extendSession) {
                window.top.sas_framework_extendSession();
            }
            if (window.top.sas_framework_extendOtherSessions) {
                window.top.sas_framework_extendOtherSessions(window.top);
            }
        },
        // Create and center a text node within a given parentNode
        createAndPlaceCenteredTextNode: function(text, parentNode) {
            var parentHeight = domStyle.getComputedStyle(parentNode).height;

            return domConstruct.place(Util.getCenteredTextNode(text, parentHeight), parentNode);
        },
        getCenteredTextNode: function(text, height) {
            var centeredNode = domConstruct.toDom(
                '<div class="stmCenteredTextWrapper">' +
                    '<div class="stmCenteredTextInner">' +
                    '<label>' + text  + '</label>' +
                    '</div>'
            );

            domStyle.set(centeredNode, 'height', height);
            return centeredNode;
        },
        // Hide the immediate children of any node
        hideImmediateChildren: function(node) {
            query('> *', node).forEach(function(nodeToHide) {
                domStyle.set(nodeToHide, 'display', 'none');
            });
        },
        // Show the immediate children of any node
        showImmediateChildren: function(node) {
            query('> *', node).forEach(function(nodeToShow) {
                domStyle.set(nodeToShow, 'display', '');
            });
        },
        // Add logic on mouseover that will add a tooltip to the cell if its text content overflows
        enableOverflowTooltipsForGrid: function(grid) {
            grid.on('.dgrid-cell:mouseover', function() {
                Util.enableOverflowTooltips(this);
            });
        },
        enableOverflowTooltipsForNode: function(node) {
            on(node, 'mouseover', function() {
                Util.enableOverflowTooltips(this);
            });
        },
        enableOverflowTooltips: function(node) {
            var title = domProp.get(node, 'title');

            if (!title) {
                if (node.offsetWidth < node.scrollWidth) {
                    domProp.set(node, 'title', domProp.get(node, 'innerText'));
                }
            }
            else {
                if (node.offsetWidth >= node.scrollWidth) {
                    domProp.set(node, 'title', null);
                }
            }
        },
        // Move cursor to the end of an INPUT or TEXTAREA
        moveCursorToEnd: function(node) {
            if (node.setSelectionRange) {
                node.selectionStart = node.selectionEnd = node.value.length;
            }
            else if (typeof node.createTextRange != 'undefined') {
                var range = node.createTextRange();
                range.collapse(false);
                range.select();
            }
        },
        // Hack to force a ValidationTextBox based widget to validate a certain way on demand
        forceValidationTextBox: function(widget, isValid) {

            widget.validator = function() {
                return isValid;
            };

            widget.validate();
        },
        forceValidationTextBoxOnce: function(widget, isValid) {
            var oldValidatorFn = widget.validator;

            widget.validator = function() {
                return isValid;
            };

            widget.validate();
            widget.validator = oldValidatorFn;
        },
        generateHelpText: function(helpText, linkLabels, helpURL) {
            var linkTemplate = '<a href="' + helpURL + '" target="_blank">${0}</a>';

            linkLabels = array.map(linkLabels, function(linkLabel) {
                return string.substitute(linkTemplate, [linkLabel]);
            });

            return string.substitute(helpText, linkLabels);
        },
        // Dojo does not provide a way to programmaticaly highlight an invalid ValidationTextBox (or descendant) while it
        // is not focused, so we have to hack into it a bit
        validateAndHighlight: function(validationWidget) {
            var isValid = validationWidget.validate();

            if (!isValid) {
                validationWidget.set('state', 'Error');
                validationWidget.set('message', validationWidget.getErrorMessage());
                validationWidget._maskValidSubsetError = false;
                validationWidget._hasBeenBlurred = true;
            }

            return isValid;
        },
        // Calculate the actual width of all the children of a node, including padding, margin, border, etc.
        // Note: this function is a bit dumb for now in that it expects all children to be in a single horizontal row
        getComputedWidthOfChildren: function(node) {
            var width = 0;

            array.forEach(node.childNodes, function(childNode) {
                width += domGeom.getMarginBox(childNode).w;
            });

            return width;
        },
        ownObserveHandle: function(owner, handle) {
            if (owner.own && typeof owner.own === 'function') {
                owner.own(
                    aspect.after(owner, 'destroy', function() {
                        handle.cancel();
                    })
                );
            }
            else {
                throw new Error(Util.NOT_DESTROYABLE_ERROR);
            }
        },
        ownObserveHandles: function(owner, handles) {
            array.forEach(handles, function(handle) {
                Util.ownObserveHandle(owner, handle);
            });
        },
        getSelectedDgridItems: function(grid) {
            var selection = grid.selection;
            var selectedItems = [];
            var id;

            if (!grid.selection) {
                throw new Error(Util.NOT_SELECTION_DGRID_ERROR);
            }

            for (id in selection) {
                selection[id] && selectedItems.push(grid.row(id).data);
            }

            return selectedItems;
        },
        isRtl: function() {
            return getComputedStyle(query("html")[0]).getPropertyValue("direction") === "rtl";
        }
    };
    
    Util.NOT_DESTROYABLE_ERROR = 'Attempted to "own" a handle using a non-destroyable Dijit';
    Util.NOT_SELECTION_DGRID_ERROR = 'The "grid" argument must be a valid dgrid with the Selection plugin';

    return Util;

});
